// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus, Florian Fischer
#pragma once

#include <atomic>
#include <chrono>

#include "CallerEnvironment.hpp"
#include "Fiber.hpp"
#include "UnboundedBlockingMpscQueue.hpp"
#include "emper.hpp"
#include "io/Future.hpp"

template <typename T>
class Actor {
 private:
	enum State {
		Stopped,
		Retrieving,
		Running,
	};

 protected:
	Runtime& runtime;

 private:
	std::atomic<State> state = {Stopped};

	UnboundedBlockingMpscQueue<T> queue;

	auto switchState(State oldState, State newState) -> bool {
		return state.compare_exchange_strong(oldState, newState, std::memory_order_acq_rel);
	}

	void setState(State newState) { state.store(newState, std::memory_order_release); }

	void actorLoop() {
		setState(Running);

		while (state.load(std::memory_order_acquire) == Running) {
			if (!switchState(Running, Retrieving)) {
				// Someone else must have set our state to Stopped
				break;
			}

			T t = queue.get([this] {
				// Prevent lost stop()
				switchState(Retrieving, Running);
			});

			receive(t);
		}
	}

 protected:
	Actor(Runtime& runtime = *Runtime::getRuntime()) : runtime(runtime), queue(runtime) {}

	virtual void receive(T t) = 0;

	void stop() { setState(Stopped); }

 public:
	virtual ~Actor() { stop(); }

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void start() {
		if (state.load(std::memory_order_acquire) != Stopped) return;

		Fiber* actorFiber = Fiber::from([this] { actorLoop(); });
		if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
			runtime.schedule(*actorFiber);
		} else {
			runtime.scheduleFromAnywhere(*actorFiber);
		}
	}

	void startFromAnywhere() { start<CallerEnvironment::ANYWHERE>(); }

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void tell(T t) {
		queue.template put<callerEnvironment>(t);
	}

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER, typename InputIt>
	void tell(InputIt begin, InputIt end) {
		queue.template put<callerEnvironment>(begin, end);
	}

	auto pendingMailboxItems() -> size_t { return queue.size(); }

	auto waitUntilIdle(long timeout) -> bool {
		const auto start = std::chrono::steady_clock::now();
		const auto deadline = start + std::chrono::seconds(timeout);
		while (!(queue.size() == 0 && state.load(std::memory_order_acquire) == Retrieving)) {
			if constexpr (emper::IO) {
				emper::sleep(1);
			} else {
				runtime.yield();
			}
			// TODO: The suppressed linter error below may be a false positive
			// reported by clang-tidy.
			// NOLINTNEXTLINE(modernize-use-nullptr)
			if (std::chrono::steady_clock::now() > deadline) {
				return false;
			}
		}

		return true;
	}
};
