// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#pragma once

#include <cassert>			// for assert
#include <cstring>			// for memset
#include <functional>		// for function
#include <iosfwd>				// for ostream
#include <type_traits>	// for remove_reference<>::type // IWYU pragma: keep
#include <utility>

#include "Common.hpp"	 // for func_t, DIE, ALIGN_TO_CACHE_LINE
#include "Debug.hpp"	 // for LOGD, LogSubsystem, LogSubsystem::C, Logger
#include "Emper.hpp"	 // for Emper::DEBUG
#include "emper-config.h"
#include "lib/math.hpp"

class AbstractFiber;
class ContextManager;
class Dispatcher;
class Fibril;

namespace em = emper::lib::math;

extern "C" [[noreturn]] void switch_and_load_context(void** toTos);
// *Not* marked as 'noreturn' because save_and_switch_context does
// *actually return at some point, unlike the other switch_*
// *functions.
extern "C" void save_and_switch_context(void** toTos, void** fromTos);
extern "C" [[noreturn]] void switch_context(void** toTos);

class EMPER_CONTEXT_ALIGNAS Context : Logger<LogSubsystem::C> {
 public:
	static constexpr size_t PAGE_SIZE = emper::ASSUME_PAGE_SIZE;

	static constexpr size_t CONTEXT_SIZE = em::roundUp(emper::MIN_CONTEXT_STACK_SIZE, PAGE_SIZE);

 private:
	static thread_local Context* currentContext;

	AbstractFiber* currentFiber = nullptr;

	void* const tos;

	void* const guardPage;

	void* const bos;

	const size_t usableStackSize;

	//	unsigned valgrindStackId;
	void* savedStackpointer;

	const void* alphaFunctionIpLocation;

	const func_t mainFunction;

	func_t startAndResumeHook;

	// NOLINTNEXTLINE(modernize-avoid-c-arrays)
	char context[CONTEXT_SIZE];

	void printTo(std::ostream& strm, bool withPtr = true) const;

	friend auto operator<<(std::ostream& strm, const Context& context) -> std::ostream&;
	friend auto operator<<=(std::ostream& strm, const Context& context) -> std::ostream&;

	friend ContextManager;

	auto getFiber() -> AbstractFiber* { return currentFiber; }

	/**
	 * The first function that a newly started context will
	 * execute. This function simply first runs the pre main hook and
	 * then the main function.
	 */
	static void kickoff() {
		Context* context = getCurrentContext();
		context->startAndResumeHook();
		context->mainFunction();
	}

	friend Dispatcher;

	// Allow Fibril to access Context::currentContext.
	friend Fibril;

	static void setCurrentFiber(AbstractFiber* fiber) {
		assert(currentContext);

		currentContext->currentFiber = fiber;
	}

	void unmap(void* sp) const;

	void maybeMarkStack();

	void recordStackUsage();

	static auto calcTos(char* context) -> void*;
	auto calcTos() -> void*;

	static auto calcGuardPage(char* context) -> void*;
	auto calcGuardPage() -> void*;

	static auto calcBos(char* context) -> void*;
	auto calcBos() -> void*;

 public:
	Context(func_t mainFunction);

	~Context();

	static auto getCurrentFiber() -> AbstractFiber* {
		assert(currentContext);

		return currentContext->currentFiber;
	}

	inline void setEmptyHook() {
		startAndResumeHook = []() {};
	}

	inline void setHook(func_t hook) { startAndResumeHook = std::move(hook); }

	[[nodiscard]] inline auto getTos() const -> const void* { return tos; }

	/**
	 * Start this context. This method is called from an unmanaged stack.
	 */
	inline void start() {
		LOGD("starting " << *this);
		currentContext = this;

		switch_context(&savedStackpointer);
	}

	/**
	 * Save the current context and switch to a new one. This method is
	 * usally used when a synchronization primitve decided that it is
	 * time to block the current execution context. In such a case
	 * 'newContext' is a free context.
	 */
	inline void saveAndStart(Context* newContext) {
		LOGD("saving and switching to " <<= *newContext);
		currentContext = newContext;

		// Save the current context and start a new one.
		save_and_switch_context(&(newContext->savedStackpointer), &savedStackpointer);
		// If we are here, then the context was unblocked.

		// Call the hook again, as it may be set by the context which
		// activated us.
		startAndResumeHook();
	}

	/**
	 * Discard the current context and restore the given one. This
	 * method is usually used when a synchronization primitive decided
	 * that it is time to unblock a context. In such a case
	 * 'newContext' is the context that became runable again. Note
	 * that this method is usually called by a so called "unblock
	 * Fiber". This method does not return.
	 */
	// The discardAndResume() method could theoretically be made static in
	// non-debug builds.
	// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
	[[noreturn]] inline void discardAndResume(Context* context) {
		//		assert(context != nullptr);

		LOGD("discarding and switching to " << context);
		currentContext = context;
		switch_and_load_context(&(context->savedStackpointer));
	}

	inline static auto getCurrentContext() -> Context* { return currentContext; }

	static auto inCurrentGuardPage(void* mem) -> bool;
	auto inGuardPage(void* mem) -> bool;
};
