// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#include "ContextManager.hpp"

#include <cassert>	// for assert
#include <new>
#include <utility>

#include "Context.hpp"
#include "Continuation.hpp"
#include "Debug.hpp"
#include "Dispatcher.hpp"
#include "Emper.hpp"
#include "Fibril.hpp"
#include "Runtime.hpp"
#include "emper-config.h"	 // // IWYU pragma: keep

class AbstractFiber;

ContextManager::ContextManager(Runtime& runtime) : MemoryManager(runtime), runtime(runtime) {}

auto ContextManager::getFreeContext() -> Context* {
	if constexpr (!emper::CONTEXT_MANAGER_WITH_MEMORY_MANAGER) {
		return new Context(this->runtime.dispatcher.getDispatchLoop());
	}

	bool malloced;
	void* memory = getMemory(&malloced);
	if (malloced) {
		return new (memory) Context(this->runtime.dispatcher.getDispatchLoop());
	}

	auto* context = static_cast<Context*>(memory);
	if constexpr (emper::STATS_STACK_USAGE == emper::StatsStackUsage::accurate) {
		// If we want accurate stack-usage stats, then we need to re-mark
		// the stack if we got the memory from the memory manager but if
		// it wasn't malloced.
		context->maybeMarkStack();
	}
	return context;
}

void ContextManager::putFreeContext(Context* context) {
	if constexpr (emper::STATS_STACK_USAGE_ENABLED) {
		context->recordStackUsage();
	}

	if constexpr (!emper::CONTEXT_MANAGER_WITH_MEMORY_MANAGER) {
		putMemory(context);
	} else {
		delete context;
	}
}

thread_local static Continuation* continuation;

void ContextManager::start() {
	Continuation originalStackContinuation;

	continuation = &originalStackContinuation;

	uintptr_t fibrilResumeValue = originalStackContinuation.setJmp();

	// We perform a longjmp with fibrilResumeValue == 0 to switch back
	// to the original stack and terminate the runtime.
	if (unlikely(!fibrilResumeValue && runtime.isTerminating())) {
		Context* currentContext = Context::currentContext;
		if (currentContext) [[likely]] {
			Context::currentContext = nullptr;
			putFreeContext(currentContext);
		}
		return;
	}

	if constexpr (emper::CONTINUATION_STEALING) {
		Fibril::tryResumeFibril(fibrilResumeValue);

		// The code below only needs to be executed if we did a long jmp
		// before. But optimizing for it is not worth it.  If we longjmp
		// and can not resume the Fibril and we where not on the initial
		// context of the Fibril, then we need to take care to recycle the
		// context.
		Context* currentContext = Context::getCurrentContext();
		if (currentContext) {
			// If we would simply call currentContext->start() here, then we
			// would jump to an invalid address. The reason is that as soon as
			// the Context::kickoff() is executed, the alpha function address,
			// which was just popped of the stack, will be potentially
			// overriden by a stack push operation. And such operations are
			// very likely. Instead we simply free the current context here,
			// so that it is re-initialized later on.

			// TODO: Adjust the assembly to not use 'ret' to reach the
			// kickoff function, but instead jmp. This would not destroy the
			// kickoff's function pointer on the stack, and we could call
			// currentContext->start() here. This would print a small
			// performance gain.
			putFreeContext(currentContext);
		}
	}

	Context* freeContext = getFreeContext();
	freeContext->start();
}

[[noreturn]] void ContextManager::switchToOriginalStack() {
	DBG("Switching to original stack from " << Context::currentContext);
	continuation->longJmp(0);
}

void ContextManager::tryResumeFibril(uintptr_t fibrilResumeValue) {
	continuation->longJmp(fibrilResumeValue);
}

/**
 * Save the current context and start a new one.
 */
void ContextManager::saveAndStartNew(func_t freshContextHook) {
	Context* freeContext = getFreeContext();
	freeContext->setHook(std::move(freshContextHook));
	Context::getCurrentContext()->saveAndStart(freeContext);
}

#if Debug > EMPER_LOG_LEVEL
#define HOOK_CAPTURE_LIST contextToFree
#else
// We only need to caputure 'this' for the LOGD() debug macro.
#define HOOK_CAPTURE_LIST this, contextToFree
#endif

/**
 * Discard the current context and resume a saved one.
 */
// The discardAndResume() method could theoretically be made static in
// non-debug builds.
// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
void ContextManager::discardAndResume(Context* context) {
	assert(context != nullptr);

	Context* contextToFree = Context::getCurrentContext();
	context->setHook([HOOK_CAPTURE_LIST] {
		LOGD("Freeing context " << contextToFree);
		putFreeContext(contextToFree);
	});

	// Since we are going to discard this context, it will never reach
	// the end of its dispatch loop, and hence we need to ensure that
	// its fiber is recycled.
	AbstractFiber* currentFiber = contextToFree->getFiber();
	runtime.dispatcher.recycle(currentFiber);

	contextToFree->discardAndResume(context);
}
