// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <cstdint>
#include <iostream>

#include "CallerEnvironment.hpp"
#include "FiberSource.hpp"
#include "emper-common.h"
#include "lib/TaggedPtr.hpp"

namespace emper {
/**
 * @brief A descriptor where to find a fiber
 *
 * A FiberHint consists of a FiberSource combined with a workerId.
 * Valid FiberSources for now are: Fiber::Source::hint{Wsq, Aq}.
 *
 * This is used during work-stealing to find work faster than stealing
 * round-robin from a random starting point.
 * The sleep strategies also use FiberHints to decide who and how to notify
 * on new work.
 */
class FiberHint {
	emper::lib::TaggedPtr _tptr;

 public:
	FiberHint() : _tptr((nullptr)) {}

	FiberHint(emper::FiberSource source) : _tptr(0, static_cast<uint16_t>(source)) {}
	FiberHint(workerid_t workerId, emper::FiberSource source)
			: _tptr(static_cast<uintptr_t>(workerId), static_cast<uint16_t>(source)) {}

	FiberHint(const FiberHint& other) : _tptr(other._tptr){};

	[[nodiscard]] auto getSource() const -> emper::FiberSource {
		return static_cast<emper::FiberSource>(_tptr.getTag());
	}

	[[nodiscard]] auto getWorker() const -> workerid_t {
		return static_cast<workerid_t>(_tptr.getRawPtrValue());
	}

	void clear() { _tptr = nullptr; }

	template <CallerEnvironment callerEnvironment>
	[[nodiscard]] static auto createNewWorkHint() -> FiberHint;

	inline operator bool() const { return _tptr; }
	inline operator void*() const { return _tptr; }
};
}	 // namespace emper

auto operator<<(std::ostream& os, const emper::FiberHint& hint) -> std::ostream&;
