// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include "Future.hpp"

#include <cassert>	// for assert
#include <cerrno>		// for errno, EAGAIN, EWOULDBLOCK
#include <ostream>	// for operator<<, ostream, basic_ost...

#include "BinaryPrivateSemaphore.hpp"	 // for BPS
#include "CallerEnvironment.hpp"			 // for CallerEnvironment
#include "Debug.hpp"									 // for LOGD, LOGW
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "Worker.hpp"
#include "io/IoContext.hpp"
#include "io/Stats.hpp"
#include "lib/TaggedPtr.hpp"

namespace emper::io {

template <CallerEnvironment callerEnvironment>
auto PartialCompletableFuture::tryComplete(int32_t res)
		-> PartialCompletableFuture::CompletionType {
	// We expect partial completions and made progress
	if (!isDependency() && partialCompletion > DISABLE_PARTIAL_COMPLETION && res > 0) {
		partialCompletion += res;
		// write/send: we have not written the whole buffer yet
		// read: we have not filled the buffer yet
		if ((size_t)partialCompletion < len) {
			IoContext* io = IoContext::getIo<callerEnvironment>();
			io->submit<callerEnvironment>(*this);
			return CompletionType::Resubmission;
		}

		return CompletionType::IncrementalCompletion;
	}

	return CompletionType::Completion;
}

// show the compiler our template incarnations
template PartialCompletableFuture::CompletionType
PartialCompletableFuture::tryComplete<CallerEnvironment::EMPER>(int32_t res);
template PartialCompletableFuture::CompletionType
PartialCompletableFuture::tryComplete<CallerEnvironment::ANYWHERE>(int32_t res);

auto Future::_wait() -> int32_t {
	LOGD("Waiting on " << this);

	sem.wait();

	return get();
}

auto Future::waitAndSetErrno() -> int32_t {
	int32_t res = wait();
	// move error codes from io_uring into errno
	if (res < 0) {
		errno = -res;
		return -1;
	}

	return res;
}

void Future::submit() {
	emper::assertInRuntime();
	assert(!isSubmitted());
	IoContext* io = IoContext::getWorkerIo();

	LOGD("submit " << this << " to IoContext " << io);

	io->submitAndReap(*this);
}

auto Future::cancel() -> int32_t {
	LOGD("Cancelling " << this);
	markCancelled();

	if (dependency) {
		dependency->cancel();
	}

	// With the separation of isSubmitted and isPrepared dependent Futures
	// are not marked as submitted anymore.
	// This means we have to check if we are canceling a not submitted Future
	// which is not a dependency.
	if ((!isSubmitted() && !isDependency()) || isForgotten()) {
		LOGW("Cancelling " << (isForgotten() ? "forgotten " : "unsubmitted ") << this);
		return -ENOENT;
	}

	if (isCompleted()) {
		return returnValue;
	}

	// Only try to cancel Futures which have actually reached the io_uring
	if (isPrepared()) {
		CancelWrapper cancellation(*this);
		// We need to cancel on a different worker
		if (!emper::IO_SINGLE_URING && submitter != IoContext::workerIo) {
			submitter->runtime.scheduleOn(*Fiber::from([&cancellation] { cancellation.submit(); }),
																		submitter->worker->getWorkerId());
		} else {
			cancellation.submit();
		}
		const int32_t cancelResult = cancellation.wait();

		if (callback) returnValue = cancelResult;
	}

	// We have to ensure that the Future can be dropped when we return.
	// This is possible when it gets completed or if we are sure it will no longer
	// be use in the IO-subsystem (SubmitActor).
	// Both can be achieved by awaiting it.
	// The former is trivial.
	// The latter is guarantied for Futures that can not be completed
	// (forgotten, callback ones) because we signal their semaphore during
	// preparation and after that they will no longer be used by EMPER.
	// The kernel though may still use resources referenced by the Future such as
	// a buffer or file descriptor.
	return _wait();
}

/**
 * Internal hiding functions for Stats::recordCompletion double dispatch
 *
 * The polymorphism functions are Future::recordCompletion and are defined in
 * Future.hpp so they are included in each compilation unit and the
 * compiler can optimize the recordCompletion calls if stats are off
 */
void Future::recordCompletionInternal(Stats& stats, int32_t res) const {
	stats.recordCompletion(*this, res);
}

void PartialCompletableFuture::recordCompletionInternal(Stats& stats, int32_t res) const {
	stats.recordCompletion(*this, res);
}

void WritevFuture::recordCompletionInternal(Stats& stats, int32_t res) const {
	stats.recordCompletion(*this, res);
}

void SendFuture::submit() {
	if constexpr (emper::IO_TRY_SYSCALL) {
		if (dependency) goto normal_submit;

		ssize_t res = send(fd, buf, len, flags | MSG_DONTWAIT);
		if (res != -1 || (errno != EAGAIN && errno != EWOULDBLOCK)) {
			markSubmitted();
			int ires = res == -1 ? -errno : static_cast<int>(res);
			complete(ires);
			return;
		}
	}

normal_submit:
	Future::submit();
}

void RecvFuture::submit() {
	if constexpr (emper::IO_TRY_SYSCALL) {
		if (dependency) goto normal_submit;

		ssize_t res = recv(fd, buf, len, flags | MSG_DONTWAIT);
		if (res != -1 || (errno != EAGAIN && errno != EWOULDBLOCK)) {
			markSubmitted();
			int ires = res == -1 ? -errno : static_cast<int>(res);
			complete(ires);
			return;
		}
	}

normal_submit:
	Future::submit();
}

void CancelWrapper::prepareSqeInternal(struct io_uring_sqe* sqe) {
	auto& future = *reinterpret_cast<Future*>(buf);
	if constexpr (!emper::IO_SINGLE_URING) {
		assert(future.submitter == IoContext::workerIo);
	}
	void* user_data = IoContext::createFutureTag(future);
	io_uring_prep_cancel(sqe, user_data, 0);
}
}	 // namespace emper::io
