// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include "io/GlobalIoContext.hpp"

#include <liburing.h>
#include <liburing/io_uring.h>
#include <sched.h>
#include <unistd.h>

#include <cassert>
#include <cerrno>
#include <ostream>
#include <vector>

#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "Worker.hpp"
#include "emper-common.h"
#include "io/Future.hpp"
#include "io/IoContext.hpp"
#include "io/SubmitActor.hpp"
#include "lib/TaggedPtr.hpp"
#include "lib/sync/Semaphore.hpp"
#include "sleep_strategy/WorkerSleepStrategy.hpp"

using emper::lib::TaggedPtr;

namespace emper::io {

static void assert_submitted(ATTR_UNUSED unsigned submitted, ATTR_UNUSED unsigned expected) {
	// Only assert how many sqe were submitted when doing the io_uring_enter syscall.
	// Reporting how many sqes were submitted uses khead which
	// is written asynchronously by the poller kthread
	if constexpr (emper::IO_SQ_POLLER == IoSqPoller::off) {
		assert(submitted == expected);
	}
}

void GlobalIoContext::completerLoop() {
	int submitted;

	if constexpr (emper::IO_WORKER_URING) {
		// wait till all workers created their IoContexts
		for (workerid_t i = 0; i < runtime.workerCount; ++i) {
			runtime.ioReadySem.wait();
		}

		LOGD("submit all worker io_uring eventfds");
		// register all eventfds for less overhead in the io_uring
		{
			LOGD("register all worker eventfds as fixed files on global io_uring");
			std::vector<int> fds;
			fds.reserve(runtime.getWorkerCount());
			for (auto* workerIo : runtime.ioContexts) {
				assert(workerIo->notificationEventFd);
				fds.push_back(workerIo->notificationEventFd);
			}

			int err = io_uring_register_files(&ring, fds.data(), fds.size());
			if (err) {
				DIE_MSG_ERRNO("registering all eventfds on the global IoCobntext failed");
			}
		}

		// prepare a read for each worker eventfd
		for (auto* workerIo : runtime.ioContexts) {
			prepareWorkerNotification(*workerIo);
		}

		// submit all eventfds
		LOGD("submit all prepared worker io_uring eventfd reads");
		submitted = io_uring_submit(&ring);
		if (unlikely(submitted < 0)) {
			DIE_MSG_ERRNO("initial global io_uring submit failed");
		}

		// We have submitted all eventfds
		assert_submitted(submitted, runtime.getWorkerCount());
	}

	// Submit the read for our termination eventfd
	struct io_uring_sqe* sqe = getSqe();
	assert(sqe);

	// We reuse the notificationEventFd member of IoContext to receive the termination notification
	io_uring_prep_read(sqe, notificationEventFd, &notificationEventFdBuf,
										 sizeof(notificationEventFdBuf), 0);
	TaggedPtr terminationEvent(static_cast<void*>(nullptr),
														 static_cast<uint16_t>(PointerTags::TerminationEvent));
	io_uring_sqe_set_data(sqe, terminationEvent);

	submitted = io_uring_submit(&ring);
	if (unlikely(submitted < 0)) {
		DIE_MSG_ERRNO("submitting termination eventfd read failed");
	}
	assert_submitted(submitted, 1);

	LOGD("start completer loop");
	while (true) {
		struct io_uring_cqe* cqe;

		// wait for completions
		int err = io_uring_wait_cqe(&ring, &cqe);
		if (unlikely(err)) {
			if (err == -EINTR) {
				continue;
			}
			errno = -err;
			DIE_MSG_ERRNO("io_uring_wait_cqe");
		}

		TaggedPtr tptr(io_uring_cqe_get_data(cqe));
		auto tag = static_cast<PointerTags>(tptr.getTag());

		// The cqe is for a IoContext.notificationEventFd read
		//  -> there are completions on this worker IoContext
		switch (tag) {
			// clang-11 does not support [[likely]] yet
			// TODO: remove the preprocessor check if clang has [[likely]] support
#if defined __has_attribute
#if __has_attribute(likely) && defined EMPER_IO_WORKER_URING
			[[likely]]	// NOLINT(clang-diagnostic-unknown-attributes)
#endif
#endif
			case PointerTags::IoContext : {
				if constexpr (emper::IO_SINGLE_URING) {
					DIE_MSG("Using a single uring there can never be IoContexts in the io_uring");
				}

				auto* worker_io = tptr.getPtr<IoContext>();
				assert(worker_io);

				// re-add the eventfd read
				prepareWorkerNotification(*worker_io);
				submitted = io_uring_submit(&ring);

				if (unlikely(submitted < 0)) {
					errno = -submitted;
					DIE_MSG_ERRNO("re-submitting eventfd read to global_ring failed");
				}

				assert_submitted(submitted, 1);

				if constexpr (emper::IO_COMPLETER_BEHAVIOR == emper::IoCompleterBehavior::wakeup) {
					// Notify the worker which's CQ has a new cqe
					workerid_t worker = worker_io->worker->getWorkerId();
					runtime.workerSleepStrategy.notifySpecific<CallerEnvironment::ANYWHERE>(worker);
				} else {
					worker_io->reapAndScheduleCompletions<CallerEnvironment::ANYWHERE>();
				}
			}
			break;

			// The cqe is for a completed Future.
			// This happens in case the global completer is dealing with a
			// partial completion.
#if defined __has_attribute
#if __has_attribute(likely) && defined EMPER_IO_SINGLE_URING
			[[likely]]	// NOLINT(clang-diagnostic-unknown-attributes)
#endif
#endif
			case PointerTags::Future : {
				auto* future = tptr.getPtr<Future>();
				LOGD("Completer got " << future << " completion");

				// Forgotten Futures can only be in the single io_uring
				if constexpr (emper::IO_SINGLE_URING) {
					if (!future) break;
				}

				int32_t res = cqe->res;

				future->recordCompletion(stats, res);
				future->completeFromAnywhere(res);
			}
			break;

			case PointerTags::Callback: {
				if constexpr (emper::IO_WORKER_URING) {
					DIE_MSG("Callbacks can never reach the GlobalIoContext from a worker IoContext");
				}

				auto* callback = tptr.getPtr<Future::CallbackInternal>();
				LOGD("Create new callback fiber for " << callback);
				auto* callbackFiber = Fiber::from(
						[&c = *callback, res = cqe->res] {
							c(res);
							delete &c;
						},
						&callback->affinity);

				runtime.scheduleFromAnywhere(*callbackFiber);
			} break;

				// someone initiated termination
			case PointerTags::TerminationEvent: {
				// TODO: Convert to LOGI once EMPER has a proper log configuration mechanism.
				LOGD("exit completer loop");
				return;
			} break;

			default:
				DIE_MSG("Unknown GlobalIoContext::PointerTag: " << tptr.getTag());
				break;
		}

		io_uring_cqe_seen(&ring, cqe);
	}
}

void GlobalIoContext::startCompleter() {
	if constexpr (emper::IO_SINGLE_URING) {
		assert(submitter);
		submitter->startFromAnywhere();
	}

	auto completerFunc = [](void* arg) -> void* {
		reinterpret_cast<GlobalIoContext*>(arg)->completerLoop();
		return nullptr;
	};
	errno = pthread_create(&completer, nullptr, completerFunc, this);

	if (unlikely(errno)) {
		DIE_MSG_ERRNO("Creating completer thread failed");
	}

	int sched_priority = 0, sched_policy;

	switch (emper::COMPLETER_SCHED_PARAM) {
		case CompleterSchedParam::unchanged:
			return;
		case CompleterSchedParam::normal:
			sched_policy = SCHED_OTHER;
			break;
		case CompleterSchedParam::nice_10:
			sched_policy = SCHED_OTHER;
			sched_priority = 10;
			break;
		case CompleterSchedParam::nice_19:
			sched_policy = SCHED_OTHER;
			sched_priority = 19;
			break;
		case CompleterSchedParam::idle:
			sched_policy = SCHED_IDLE;
			break;
	}

	const struct sched_param param = {sched_priority};
	errno = pthread_setschedparam(completer, sched_policy, &param);

	if (unlikely(errno)) {
		DIE_MSG_ERRNO("Setting scheduler parameter on completer thread failed");
	}
}

void GlobalIoContext::initiateTermination() {
	uint64_t write_buf = 1;
	ssize_t bytes_written = write(notificationEventFd, &write_buf, sizeof(write_buf));
	if (bytes_written != sizeof(write_buf)) {
		DIE_MSG_ERRNO("Writing to globalIo termination eventfd failed");
	}
}

void GlobalIoContext::waitUntilFinished() const {
	errno = pthread_join(completer, nullptr);
	if (unlikely(errno)) {
		DIE_MSG_ERRNO("pthread_join failed for the globalCompleter");
	}
}

void GlobalIoContext::prepareWorkerNotification(IoContext& workerIo) {
	struct io_uring_sqe* sqe = getSqe();

	io_uring_prep_read(sqe, workerIo.worker->getWorkerId(), &workerIo.notificationEventFdBuf,
										 sizeof(workerIo.notificationEventFdBuf), 0);
	// set flags after io_uring_prep* calls because they will overwrite the flags
	sqe->flags |= IOSQE_FIXED_FILE;
	io_uring_sqe_set_data(sqe, TaggedPtr(&workerIo, static_cast<uint16_t>(PointerTags::IoContext)));
}
}	 // namespace emper::io
