// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2022 Florian Fischer
#include "ShardedFileBuffer.hpp"

#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>

#include <atomic>
#include <cassert>

#include "Common.hpp"
#include "Runtime.hpp"
#include "emper-common.h"
#include "io.hpp"

namespace emper::lib {

std::atomic<size_t> ShardedFileBuffer::ids = 0;
thread_local ShardedFileBuffer::WorkerBuffers* ShardedFileBuffer::workerBuffers = nullptr;

std::mutex ShardedFileBuffer::registerLock;
ShardedFileBuffer::RuntimeBuffers ShardedFileBuffer::buffersPerRuntime;

ShardedFileBuffer::ShardedFileBuffer(const Runtime& runtime, int fd) : runtime(runtime), fd(fd) {
	std::stringstream sst;
	sst << "fd: " << fd;
	file = sst.str();

	id = ids.fetch_add(1);
}

ShardedFileBuffer::ShardedFileBuffer(const Runtime& runtime, const std::string& file)
		: runtime(runtime), file(file) {
	fd = open(file.c_str(), O_RDWR | O_TRUNC | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		DIE_MSG_ERRNO("opening file " << file << " failed");
	}

	id = ids.fetch_add(1);
}

ShardedFileBuffer::~ShardedFileBuffer() {
	flush();
	close(fd);
}

void ShardedFileBuffer::flush() {
	std::stringstream content;
	const auto it = buffersPerRuntime.find(&runtime);
	assert(it != buffersPerRuntime.end());
	const Shards* shards = it->second;
	for (const auto* shard : *shards) {
		const auto it = shard->find(id);
		if (it != shard->end()) content << it->second.str();
	}

	const std::string& out = content.str();
	if (emper::io::writeFileAndWait(fd, out.c_str(), out.length(), -1) == -1)
		DIE_MSG_ERRNO("flushing sharded file buffer to " << file << " failed");
}

void ShardedFileBuffer::registerRuntime(Runtime& runtime) {
	std::lock_guard<std::mutex> lock(registerLock);

	assert(buffersPerRuntime.find(&runtime) == buffersPerRuntime.end());

	auto* shards = new Shards(runtime.getWorkerCount());
	buffersPerRuntime.emplace(&runtime, shards);

	runtime.addNewWorkerHook([&runtime](workerid_t workerId) {
		workerBuffers = new WorkerBuffers();
		const auto it = buffersPerRuntime.find(&runtime);
		assert(it != buffersPerRuntime.end());

		auto& shards = *(it->second);
		shards[workerId] = workerBuffers;
	});
}

void ShardedFileBuffer::unregisterRuntime(const Runtime& runtime) {
	std::lock_guard<std::mutex> lock(registerLock);

	const auto it = buffersPerRuntime.find(&runtime);
	assert(it != buffersPerRuntime.end());
	auto* shards = it->second;
	for (auto* shard : *shards) {
		delete shard;
	}
	delete shards;

	buffersPerRuntime.erase(it);
}
}	 // namespace emper::lib
