// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus, Nicolas Pfeiffer
#pragma once

#include <deque>
#include <mutex>
#include <type_traits>

#include "Common.hpp"
#include "StealingResult.hpp"

// Forward declaration, declared and defined in Fibril.hpp.
namespace emper {
void maybeLockFibril(void* queueItem);
}

namespace adt {

template <typename I, const uintptr_t SIZE>
class LockedQueue {
 private:
	std::mutex queue_mutex;

	std::deque<I> deque;

 public:
	// We can not mark this method 'const' as std::lock_guard modifies
	// the mutex.
	auto isFull() -> bool {
		std::lock_guard<std::mutex> lock(queue_mutex);
		return deque.size() == SIZE;
	}

	auto freeSlots() -> size_t {
		std::lock_guard<std::mutex> lock(queue_mutex);
		return SIZE - deque.size();
	}

	auto usedSlots() -> size_t {
		std::lock_guard<std::mutex> lock(queue_mutex);
		return deque.size();
	}

	auto pushBottom(const I item) -> bool {
		std::lock_guard<std::mutex> lock(queue_mutex);

		if (deque.size() == SIZE) return false;

		deque.push_back(item);

		return true;
	}

	template <const int maxRetries>
	auto popTop(I* itemPtr) -> emper::StealingResult {
		std::lock_guard<std::mutex> lock(queue_mutex);

		if (deque.empty()) return emper::StealingResult::Empty;

		*itemPtr = deque.front();

		// This is one of the ugliest places in EMPER. Here we specialize
		// the LockedQueue for locked continuation stealing, where it is
		// important that the Fibril gets locked *before* the queue lock
		// is released. For more information, see the IPDPS 2021 paper
		// by Schmaus et al. [schmaus2021nowa], Figure 6. and Listing 2.
		if constexpr (emper::CONTINUATION_STEALING_MODE_LOCKED &&
									std::is_same<AbstractFiber*, I>::value) {
			emper::maybeLockFibril(*itemPtr);
		}

		deque.pop_front();

		return emper::StealingResult::Stolen;
	}

	auto popBottom(I* itemPtr) -> bool {
		std::lock_guard<std::mutex> lock(queue_mutex);

		if (deque.empty()) return false;

		*itemPtr = deque.back();

		deque.pop_back();

		return true;
	}
};
}	 // namespace adt
