// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021-2022 Florian Schmaus
#include "Worker.hpp"

#include <algorithm>
#include <string>

#include "Emper.hpp"
#include "lib/util.hpp"

namespace eu = emper::lib::util;

namespace emper::stats {

thread_local Worker* worker = nullptr;

Worker::Worker(workerid_t workerCount)
		: unblockAffinitiesGeneric(std::vector<uint32_t>(workerCount)),
			unblockAffinitiesIo(std::vector<uint32_t>(workerCount)),
			lastStackUsage(32) {}

void Worker::recordBlockedContext(size_t blockedContextCount, BlockablePurpose blockablePurpose) {
	maxBlockedContexts = std::max(maxBlockedContexts, blockedContextCount);

	switch (blockablePurpose) {
		case BlockablePurpose::GENERIC:
			blockedGeneric++;
			break;
		case BlockablePurpose::IO:
			blockedIo++;
			break;
	}
}

void Worker::recordUnblockedContext(workeraffinity_t workerAffinity,
																		BlockablePurpose blockablePurpose) {
	switch (blockablePurpose) {
		case BlockablePurpose::GENERIC:
			unblockAffinitiesGeneric[workerAffinity]++;
			break;
		case BlockablePurpose::IO:
			unblockAffinitiesIo[workerAffinity]++;
			break;
	}
}

void Worker::recordStackUsage(size_t stackUsage) {
	maxStackUsage = std::max(stackUsage, maxStackUsage);
	minStackUsage = std::min(stackUsage, minStackUsage);
	averageStackUsage.update(stackUsage);
	lastStackUsage.push_back(stackUsage);
}

auto Worker::operator+=(const Worker& other) -> Worker& {
	maxBlockedContexts = std::max(maxBlockedContexts, other.maxBlockedContexts);
	blockedGeneric += other.blockedGeneric;
	blockedIo += other.blockedIo;

	for (decltype(unblockAffinitiesGeneric)::size_type i = 0; i < unblockAffinitiesGeneric.size();
			 i++) {
		unblockAffinitiesGeneric[i] += other.unblockAffinitiesGeneric[i];
		unblockAffinitiesIo[i] += other.unblockAffinitiesIo[i];
	}

	if constexpr (emper::STATS_STACK_USAGE_ENABLED) {
		maxStackUsage = std::max(maxStackUsage, other.maxStackUsage);
		minStackUsage = std::min(minStackUsage, other.minStackUsage);
		averageStackUsage.update(other.averageStackUsage.getAverage());

		// We clear the history of last stack usages, as "adding" them
		// together is not sensible.
		lastStackUsage.clear();
	}

	return *this;
}

void Worker::print(std::ostream& out) {
	out << "max-blocked-contexts: " << std::to_string(maxBlockedContexts) << std::endl
			<< "blocked-generic: " << std::to_string(blockedGeneric) << std::endl;
	for (decltype(unblockAffinitiesGeneric)::size_type i = 0; i < unblockAffinitiesGeneric.size();
			 ++i) {
		out << "unblock-affinities-generic" << std::to_string(i) << ": "
				<< std::to_string(unblockAffinitiesGeneric[i]) << std::endl;
	}
	out << "blocked-io: " << std::to_string(blockedIo) << std::endl;
	for (decltype(unblockAffinitiesIo)::size_type i = 0; i < unblockAffinitiesIo.size(); ++i) {
		out << "unblock-affinities-io" << std::to_string(i) << ": "
				<< std::to_string(unblockAffinitiesIo[i]) << std::endl;
	}

	if constexpr (emper::STATS_STACK_USAGE_ENABLED) {
		eu::writeByteStats("max-stack-usage", maxStackUsage, out);
		eu::writeByteStats("min-stack-usage", minStackUsage, out);
		auto averageStackUsageInteger = static_cast<uintptr_t>(averageStackUsage.getAverage());
		eu::writeByteStats("avg-stack-usage", averageStackUsageInteger, out);
		if (!lastStackUsage.empty()) {
			out << "last-stack-usage: ";
			for (size_t stackUsage : lastStackUsage) {
				eu::writeByteStats(stackUsage, out);
			}
			out << std::endl;
		}
	}
}

}	 // namespace emper::stats
